/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * =====================================================================================
 *
 *       Filename:  frequency_measure.cpp
 *
 *    Description:  Class for automatic measure frequency
 *
 *
 *        Version:  1.0
 *        Created:  31.01.2012 13:24:53
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Lucjan Bryndza (lb), lucck@boff.pl
 *   Organization:
 *
 * ===================================================================================== */

#include "freqency_measure.hpp"
#include <system.h>
#include <dbglog.h>
/* ------------------------------------------------------------------ */
namespace app {
/* ------------------------------------------------------------------ */
//Internal unnamed namespace
namespace
{
	//ADC divider on/off
	GPIO_TypeDef * const FREQ_PORT = GPIOA;
	const unsigned FREQ_SEL1_PIN = 4;
	const unsigned FREQ_SEL2_PIN = 5;
	const unsigned FREQ_PWR 	 = 3;
}

/* ------------------------------------------------------------------ */
//Constructor
freqency_measure::freqency_measure()
	: m_freq_meter( TIM3 ), m_mode(mode_invalid), m_prev_freq(0), m_idle(false)
{
	//Configure control gpios
	stm32::io_config(FREQ_PORT , FREQ_SEL1_PIN, stm32::GPIO_MODE_10MHZ, stm32::GPIO_CNF_GPIO_PP);
	stm32::io_config(FREQ_PORT , FREQ_SEL2_PIN, stm32::GPIO_MODE_10MHZ, stm32::GPIO_CNF_GPIO_PP);
	stm32::io_config(FREQ_PORT , FREQ_PWR, stm32::GPIO_MODE_10MHZ, stm32::GPIO_CNF_GPIO_PP );
	//OFF device by default
	stm32::io_clr( FREQ_PORT , FREQ_PWR );
	//Switch mode to divide 64
	set_mode( mode_divide_64 );
}
/* ------------------------------------------------------------------ */
//Set device in selected mode
void freqency_measure::set_mode( int mode )
{
	if( static_cast<periph_mode>(mode) == m_mode ) return;
	switch( mode )
	{
	case mode_direct:	//Without divider
		stm32::io_clr( FREQ_PORT , FREQ_SEL1_PIN );
		stm32::io_set( FREQ_PORT , FREQ_SEL2_PIN );
		break;
	case mode_divide_64:	//With divider 64
		stm32::io_set( FREQ_PORT , FREQ_SEL1_PIN );
		stm32::io_clr( FREQ_PORT , FREQ_SEL2_PIN );
		break;
	default:
		break;
	}
	m_mode = static_cast<periph_mode>(mode);
}
/* ------------------------------------------------------------------ */
//Enable or disable perhipheral
void freqency_measure::enable( bool en )
{
	if( en )
		stm32::io_set( FREQ_PORT , FREQ_PWR );
	else
		stm32::io_clr( FREQ_PORT , FREQ_PWR );
}
/* ------------------------------------------------------------------ */
//Destructor
freqency_measure::~freqency_measure()
{
	//Destructor setup control gpios to the initial state
	stm32::io_config( FREQ_PORT , FREQ_SEL1_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT );
	stm32::io_config( FREQ_PORT , FREQ_SEL2_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT );
	stm32::io_config( FREQ_PORT , FREQ_PWR, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT );
	dbprintf("Destructor called");
}
/* ------------------------------------------------------------------ */
//Get calculated value
int freqency_measure::operator()(real_t &mvalue)
{
	unsigned long curr_f = m_freq_meter();
	if( curr_f == dev::tim_freq_measure::NO_NEW_DATA )
	{
		return RESULT_NO_NEWDATA;
	}
	curr_f =  m_mode==mode_direct?curr_f:curr_f*HW_DIVIDE;
	mvalue = curr_f;
	update_idle_state( curr_f );
	return  RESULT_OK;
}

/* ------------------------------------------------------------------ */
} /* namespace app */
/* ------------------------------------------------------------------ */
