/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * keyboard.cpp
 *
 *  Created on: 26-01-2012
 *      Author: lucck
 */
/* ------------------------------------------------------------------ */
#include "keyboard.hpp"
#include "sys_time.hpp"
#include <system.h>
#include <dbglog.h>
#include <bitops.hpp>
#include <cstddef>
/* ------------------------------------------------------------------ */
namespace dev {
/* ------------------------------------------------------------------ */
namespace
{
	template <typename T> inline T bv(T b)
    {
    	return 1<<b;
    }
    static const unsigned ONOFF_PIN = 1;
	static const unsigned NEXT_PIN = 2;
	GPIO_TypeDef * const KEY_PORT = GPIOA;

	inline void init_keys()
	{
		using namespace stm32;
		io_config( KEY_PORT, ONOFF_PIN, GPIO_MODE_INPUT, GPIO_CNF_IN_PULLUP);
		io_config( KEY_PORT, NEXT_PIN, GPIO_MODE_INPUT, GPIO_CNF_IN_PULLUP);
	}
	inline void deinit_keys()
	{
			using namespace stm32;
			io_config( KEY_PORT, ONOFF_PIN, GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT);
			io_config( KEY_PORT, NEXT_PIN, GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT);
	}

	inline unsigned key_inp()
	{
		using namespace stm32;
		return (io_get(KEY_PORT,ONOFF_PIN) |
			    (io_get(KEY_PORT,NEXT_PIN)<<1)) & 0x03;
	}
	//Keyboard handler
	keyboard * p_kbd_obj;
}

/* ------------------------------------------------------------------ */
//Default constructor
keyboard::keyboard()
	: repeat_k(0), delay_k(0),  bitset_cycle(0)
{
	for( size_t s = 0; s<sizeof(keys_bitset); s++)
		keys_bitset[s] = 0;
	if( !p_kbd_obj )
	{
		//Initialize kbd
		init_keys();
		p_kbd_obj = this;
	}
}
/* ------------------------------------------------------------------ */
//Destructor
keyboard::~keyboard()
{
	deinit_keys();
	dbprintf("Destructor called");
}
/* ------------------------------------------------------------------ */
void keyboard::got_key( key_val_t port_key )
{
	if(++repeat_k == REPEAT_TIME)
	{
		repeat_k = 0;
		if(port_key)
		{
			if(port_key == last_key)
			{
				if(!delay_k)
				{
					key_back = port_key | KEY_RPT;
				}
				else
				{
					--delay_k;
				}
			}
			else
			{
				last_key = port_key;
				delay_k = DELAY_TIME;
			}
		}
		else
		{
			if( last_key && delay_k )
				key_back = last_key;
			last_key = 0;
		}
	}
	if( key_back )
	{
		if( stm32::atomic_try_write_byte( &key, key_back ) == 0)
			key_back = 0;
	}
}
/* ------------------------------------------------------------------ */
void keyboard::on_time_event()
{
	if( bitset_cycle < NUM_CYCLES )
	{
		for( size_t nk = 0; nk < NUM_KEYS; nk++ )
		{
			keys_bitset[ nk ] |= ((key_inp()>>nk) & 1) << bitset_cycle;
		}
		bitset_cycle++;
	}
	else
	{
		key_val_t key = 0;
		for( size_t nk = 0; nk < NUM_KEYS; nk++ )
		{
			key |= (fnd::ones32( keys_bitset[ nk ])>=NUM_ONES)?(1<<nk):(0);
		}
		bitset_cycle = 0;
		for( size_t s = 0; s<sizeof(keys_bitset); s++)
			keys_bitset[s] = 0;
		got_key( key );
	}
}
/* ------------------------------------------------------------------ */
//Get keys
int keyboard::get() const
{
	return stm32::atomic_xchg_byte( &key, 0 );
}
/* ------------------------------------------------------------------ */
//Wait for all keys will be released
int keyboard::wait_for_release_all() const
{
	for(int n=0; n<6; ++n)
	{
		tim::timer_wait( 20 );
		if( key_inp() ) n = 0;
	}
	return get();
}
/* ------------------------------------------------------------------ */
//Interrupt
namespace tim {
namespace _internal {

	//Keyboard tick handler
	void keyboard_tick_handler()
	{
		if( p_kbd_obj )
			p_kbd_obj->on_time_event();
	}

}}
/* ------------------------------------------------------------------ */
} /* namespace app */


/* ------------------------------------------------------------------ */
