/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * =====================================================================================
 *
 *       Filename:  lc_measure.cpp
 *
 *    Description:  Class for automatic measure L or C
 *
 *
 *        Version:  1.0
 *        Created:  31.01.2012 13:24:53
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Lucjan Bryndza (lb), lucck@boff.pl
 *   Organization:
 *
 * ===================================================================================== */

#include "lc_measure.hpp"
#include <cmath>
#include <dbglog.h>
#include <stm32pwr.h>
#include <stm32bkp.h>
#include <stm32rcc.h>
/* ------------------------------------------------------------------ */
namespace app {
/* ------------------------------------------------------------------ */
//Internal unnamed namespace
namespace
{
	//ADC divider on/off
	GPIO_TypeDef * const LC_PORT = GPIOB;
	const unsigned LM_SEL_PIN	= 11;
	const unsigned CM_SEL_PIN	= 15;
	const unsigned CAL_SEL_PIN	= 12;
	const unsigned LC_PWR_PIN	= 10;
	//Enum incrementer
	template <typename T> inline T& operator++( T& v )
	{
		v = static_cast<T>(static_cast<int>(v) + 1);
		return v;
	}
    const unsigned short backup_ram_key = 0x1979;
	const int backup_min_val = 10000;
	const int backup_max_val = 2000000;
    const unsigned short backup_L_flag = 0x8000;
    const unsigned short backup_C_flag = 0x4000;
    const unsigned short backup_key_mask = 0x3fff;
}

/* ------------------------------------------------------------------ */
//Constructor
lc_measure::lc_measure()
	: m_freq_meter(TIM2),  m_num_cycles(0), m_prev_freq(0), m_mode( mode_C )
	  , m_state( s_switch_C ), m_f1_L(0), m_f2_L(0), m_f1_C(0), m_f2_C(0),
	    m_idle( false )
{
	using namespace stm32;
	rcc_periph_clock_cmd( rcc_clk_apb1,  RCC_APB1Periph_BKP, true );
	pwr_backup_access_cmd( true );
	//Configure control GPIOS
	io_config(LC_PORT, LM_SEL_PIN,  GPIO_MODE_10MHZ,  GPIO_CNF_GPIO_OD);
	io_config(LC_PORT, CM_SEL_PIN,  GPIO_MODE_10MHZ,  GPIO_CNF_GPIO_OD);
	io_config(LC_PORT, CAL_SEL_PIN, GPIO_MODE_10MHZ,  GPIO_CNF_GPIO_OD);
	io_config(LC_PORT, LC_PWR_PIN,  GPIO_MODE_10MHZ,  GPIO_CNF_GPIO_OD);
	io_set( LC_PORT, LC_PWR_PIN );	//Off state
	switch_LCC( dmode_C );
	//Restore backup data
    const unsigned short key_reg = bkp_read_backup_register( BKP_DR1 );
	if( ( key_reg & backup_key_mask) == backup_ram_key )
	{
		int tmp;
		//Try to restore f1_L
		tmp = (static_cast<int>(bkp_read_backup_register( BKP_DR2 )) << 16) | bkp_read_backup_register( BKP_DR3 );
		if(tmp >= backup_min_val && tmp <= backup_max_val && (key_reg & backup_L_flag))
		{
			m_f1_L = tmp;
		}
		else
		{
			dbprintf("Invalid value m_f1_L %d", tmp);
		}
		//Try to restore f2_L
		tmp = (static_cast<int>(bkp_read_backup_register( BKP_DR4 )) << 16) | bkp_read_backup_register( BKP_DR5 );
		if(tmp >= backup_min_val && tmp <= backup_max_val && (key_reg & backup_L_flag))
		{
			m_f2_L = tmp;
		}
		else
		{
			dbprintf("Invalid value m_f2_L %d", tmp);
		}
		//Try to restore f1_C
		tmp = (static_cast<int>(bkp_read_backup_register( BKP_DR6 )) << 16) | bkp_read_backup_register( BKP_DR7 );
		if(tmp >= backup_min_val && tmp <= backup_max_val && (key_reg & backup_C_flag))
		{
			m_f1_C = tmp;
		}
		else
		{
			dbprintf("Invalid value %d m_f1_C", tmp);
		}
		//Try to restore f2_C
		tmp = (static_cast<int>(bkp_read_backup_register( BKP_DR8 )) << 16) | bkp_read_backup_register( BKP_DR9 );
		if(tmp >= backup_min_val && tmp <= backup_max_val && (key_reg & backup_C_flag))
		{
			m_f2_C = tmp;
		}
		else
		{
			dbprintf("Invalid value %d  m_f1_C", tmp);
		}
	}
}
/* ------------------------------------------------------------------ */
//Destructor
lc_measure::~lc_measure()
{
	using namespace stm32;
	enable(false);
	//Deconfigure gpios
	io_config(LC_PORT, LM_SEL_PIN,  GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT);
	io_config(LC_PORT, CM_SEL_PIN,  GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT);
	io_config(LC_PORT, CAL_SEL_PIN, GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT);
	io_config(LC_PORT, LC_PWR_PIN,  GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT);
	//Try to save data to backup regs
    unsigned short key_reg = backup_ram_key;
	if( m_f1_L > 0 && m_f2_L > 0)
    {
        dbprintf("backup v1=%d",  m_f1_L);
        bkp_write_backup_register( BKP_DR2, m_f1_L >> 16 );
	    bkp_write_backup_register( BKP_DR3, m_f1_L );
	    dbprintf("backup v2=%d",  m_f2_L);
	    bkp_write_backup_register( BKP_DR4, m_f2_L >> 16 );
	    bkp_write_backup_register( BKP_DR5, m_f2_L );
        key_reg |= backup_L_flag;
    }
    if( m_f1_C > 0 && m_f2_C > 0)
    {
        dbprintf("backup v3=%d",  m_f1_C);
        bkp_write_backup_register( BKP_DR6, m_f1_C >> 16 );
        bkp_write_backup_register( BKP_DR7, m_f1_C );
	    dbprintf("backup v3=%d",  m_f2_C);
	    bkp_write_backup_register( BKP_DR8, m_f2_C >> 16 );
	    bkp_write_backup_register( BKP_DR9, m_f2_C );
        key_reg |= backup_C_flag;
    }
	bkp_write_backup_register( BKP_DR1, key_reg );
	stm32::pwr_backup_access_cmd( false );
	dbprintf("Destructor called");
}
/* ------------------------------------------------------------------ */
//Get calculated value
int lc_measure::operator()(real_t &mvalue)
{
    static const int C_num_stable_cycles = 12;
    static const int C_inital_wait_cycles = 4;
    static const int C_f_comp_divide =  dev::tim_freq_measure::RAW_DATA_DIVIDE * 2;
    static const int C_f_percent_per_stage = 33;
	const unsigned long curr_f = m_freq_meter();
	if( curr_f == dev::tim_freq_measure::NO_NEW_DATA )
	{
		return RESULT_NO_NEWDATA;
	}
	//dbprintf("state %d mode %d freq %d", m_state, m_mode, curr_f);
	if( m_mode == mode_calibrate_L || m_mode == mode_calibrate_C )
	{
	    m_idle = false;
		if(curr_f < C_invalid_freq_value )
	    {
	    	m_prev_freq = curr_f;
	    	return RESULT_INVALID_CALIBRATION;
	    }
        switch( m_state )
		{
		case s_switch_C:
		{
			short tmp_cycles;
			switch_LCC( m_mode==mode_calibrate_C?dmode_C:dmode_L );
			if( ++m_num_cycles > C_inital_wait_cycles )
			{
				m_state = s_wait_C;
				tmp_cycles = m_num_cycles;
				m_num_cycles = 0;
				m_tmp_f1 = 0;
			}
			else
			{
				tmp_cycles = m_num_cycles;
			}
			mvalue = curr_f;
			m_prev_freq = curr_f;
			return RESULT_CALIBRATE_0PERCENT + (33*tmp_cycles)/C_inital_wait_cycles;
		}
		case s_wait_C:
		{
			if( m_prev_freq/C_f_comp_divide  == curr_f/C_f_comp_divide) ++m_num_cycles;
			else m_num_cycles = 0;
			short tmp_cycles = m_num_cycles;
			if( m_num_cycles >= C_num_stable_cycles)
			{
				m_state = s_measure_C;
				m_num_cycles = 0;
			}
			m_prev_freq = curr_f;
			mvalue = curr_f;
			return RESULT_CALIBRATE_0PERCENT +
					(C_f_percent_per_stage*tmp_cycles)/C_num_stable_cycles + C_f_percent_per_stage;
		}
		case s_measure_C:
		{
			m_tmp_f1 = curr_f;
			m_state = s_wait_CC;
			switch_LCC( m_mode==mode_calibrate_C?dmode_C_Cref:dmode_L_Cref );
			mvalue = curr_f;
			m_prev_freq = curr_f;
			return RESULT_CALIBRATE_0PERCENT + 2*C_f_percent_per_stage;
		}
		case s_wait_CC:
		{
			if( m_prev_freq/C_f_comp_divide == curr_f/C_f_comp_divide) ++m_num_cycles;
			else m_num_cycles = 0;
			if( m_num_cycles >= C_num_stable_cycles)
				m_state = s_measure_CC;
			m_prev_freq = curr_f;
			mvalue = curr_f;
			return RESULT_CALIBRATE_0PERCENT +
					(C_f_percent_per_stage*m_num_cycles)/C_num_stable_cycles + 2*C_f_percent_per_stage;
		}
		case s_measure_CC:
		{
			m_state = s_ready;
			if( m_mode == mode_calibrate_C )
			{
				switch_LCC( dmode_C );
				m_mode = mode_C;
				m_f1_C = m_tmp_f1;
				m_f2_C = curr_f;
				dbprintf("Calibration C finished f1=%d f2=%d", m_f1_C, m_f2_C);
			}
			else if( m_mode == mode_calibrate_L  )
			{
				switch_LCC( dmode_L );
				m_mode = mode_L;
				m_f1_L = m_tmp_f1;
				m_f2_L = curr_f;
				dbprintf("Calibration L finished f1=%d f2=%d", m_f1_L, m_f2_L);
			}
			mvalue = curr_f;
			m_prev_freq = curr_f;
			return RESULT_CALIBRATE_100PERCENT;
		}
		case s_ready:
			break;
		}
	}
	else if( m_mode == mode_L )	    //Mode L
	{
		update_idle_state( curr_f );
		if( m_f1_L == 0 || m_f2_L == 0)
		{
			m_state = s_switch_C;
			m_prev_freq = curr_f;
			return RESULT_NO_CALIBRATION;
		}
		else if(curr_f < C_invalid_freq_value )
		{
			m_prev_freq = curr_f;
			return RESULT_INVALID;
		}
		const real_t F1_F3 = static_cast<real_t>(m_f1_L)/static_cast<real_t>(curr_f);
		const real_t F1_F2 = static_cast<real_t>(m_f1_L)/static_cast<real_t>(m_f2_L);
		const real_t PI = std::atan(1.0)*4;
		const real_t C_Cal_1 = 1.0/ get_Cref_value();
		const real_t PI_F1 = 1.0/( 2.0 * PI * static_cast<real_t>(m_f1_L) );
		const real_t Lx = ( F1_F3*F1_F3 -1 ) * ( F1_F2*F1_F2 -1) * C_Cal_1 * ( PI_F1*PI_F1 );
		//dbprintf("Value L is %d", (int)(Lx * 10E7));
		mvalue = Lx;
	}
	else if( m_mode == mode_C )	//Mode C
	{
		update_idle_state( curr_f );
		if( m_f1_C == 0 || m_f2_C == 0)
		{
			m_state = s_switch_C;
			m_prev_freq = curr_f;
			return RESULT_NO_CALIBRATION;
		}
		else if(curr_f < C_invalid_freq_value )
		{
			m_prev_freq = curr_f;
			return RESULT_INVALID;
		}
		const real_t F1_F3 = static_cast<real_t>(m_f1_C)/static_cast<real_t>(curr_f);
		const real_t F1_F2 = static_cast<real_t>(m_f1_C)/static_cast<real_t>(m_f2_C);
		const real_t C_Cal = get_Cref_value();
		const real_t Cx = (F1_F3*F1_F3 - 1)/(F1_F2*F1_F2 - 1) * C_Cal;
		//dbprintf("Value C is %d F1 %d F2 %d", (int)(Cx * 1E13), m_f1, m_f2);
		mvalue = Cx;
	}
	m_prev_freq = curr_f;
	return RESULT_OK;

}
/* ------------------------------------------------------------------ */
//Enable or disable perhipheral
void lc_measure::enable( bool en )
{
	if( en )
		stm32::io_clr( LC_PORT, LC_PWR_PIN );
	else
		stm32::io_set( LC_PORT, LC_PWR_PIN );
}
/* ------------------------------------------------------------------ */
//Switch L, C, C+Cref
void lc_measure::switch_LCC( device_mode dev)
{
	switch( dev )
	{
	case dmode_L:	//Measure L
		stm32::io_clr( LC_PORT, CAL_SEL_PIN );
		stm32::io_set( LC_PORT , LM_SEL_PIN );
		stm32::io_clr( LC_PORT , CM_SEL_PIN );
		break;
	case dmode_C:	//Measure C
		stm32::io_clr( LC_PORT, CAL_SEL_PIN );
		stm32::io_clr( LC_PORT , LM_SEL_PIN );
		stm32::io_set( LC_PORT , CM_SEL_PIN );
		break;
	case dmode_C_Cref:	//Calibrate (C + C)
		stm32::io_set( LC_PORT, CAL_SEL_PIN );
		stm32::io_clr( LC_PORT , LM_SEL_PIN );
		stm32::io_set( LC_PORT , CM_SEL_PIN );
		break;
	case dmode_L_Cref:
		stm32::io_set( LC_PORT, CAL_SEL_PIN );
		stm32::io_set( LC_PORT , LM_SEL_PIN );
		stm32::io_clr( LC_PORT , CM_SEL_PIN );
	default:
		break;
	}
}
/* ------------------------------------------------------------------ */
	//Set device in selected mode
void lc_measure::set_mode( int mode )
{
	if(m_mode == mode_C || m_mode == mode_L || m_mode == mode_calibrate_L || m_mode == mode_calibrate_C )
	{
		m_mode = static_cast<periph_mode>(mode);
		if(m_mode == mode_C)
			switch_LCC( dmode_C );
		else if (m_mode == mode_L)
			switch_LCC( dmode_L );
	}
	if(m_mode == mode_calibrate_L || m_mode == mode_calibrate_C )
	{
		m_state = s_switch_C;
	}
	dbprintf("mode set mode=%d m_mode=%d, m_state=%d", mode, m_mode, m_state );
}
/* ------------------------------------------------------------------ */
//Get cref value
real_t lc_measure::get_Cref_value() const
{
	const uint32_t FlashBaseAddr = 0x08000000;
	const uint32_t kB = 1024;
	const uint32_t Cref_addr = FlashBaseAddr + stm32::flash_mem_size()*kB - sizeof(uint32_t);
	const real_t C_Cal_def = 2.2E-9;	//2n2
	const real_t C_unit_factor = 1E-14;	//Int is stored 1n = 100000
	const int int_C =  *(reinterpret_cast<volatile int32_t*>( Cref_addr ));
	if( int_C >= Cref_int_min  && int_C<= Cref_int_max )
		return static_cast<real_t>(int_C) * C_unit_factor;
	else
		return C_Cal_def;
}
/* ------------------------------------------------------------------ */
} /* namespace app */
/* ------------------------------------------------------------------ */
