/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * lcddisp.cpp
 *
 *  Created on: 08-01-2012
 *      Author: lucck
 */
/* ------------------------------------------------------------------ */
#include "lcddisp.h"
#include "stm32f10x_lib.h"
#include "sys_time.hpp"
#include <dbglog.h>
/* ------------------------------------------------------------------ */
namespace dev {
/* ------------------------------------------------------------------ */
namespace		//Private namespace
{
/* ------------------------------------------------------------------ */
	GPIO_TypeDef * const LCD_PORT = GPIOB; //Port Display port def
	const unsigned LCD_RS_PIN = 1<<2;	//LCD RS signal Pin9
	const unsigned LCD_RW_PIN = 1<<1;  //LCD RW signal Pin10
	const unsigned LCD_EN_PIN = 1<<0;  //LCD RW signal Pin11
	const unsigned LCD_ON_PIN = 9;
	const unsigned LCD_DATA_BITS_MASK = 0x01E0;
	const unsigned LCD_GPIO_DATA_CRL_MASK = 0xfff00000;
	const unsigned LCD_GPIO_DATA_CRH_MASK = 0x0000000f;
	const unsigned LCD_GPIO_ALL_CRL_MASK = 0xfff00fff;
	const unsigned LCD_GPIO_ALL_CRH_MASK = 0x0000000f;

/* ------------------------------------------------------------------ */
//Return CRHL gpio reg at pos
static inline unsigned CRL_VAL( unsigned mode, unsigned conf, unsigned pos)
{
	return (mode | (conf << 2))  << (4 * pos);
}
static inline unsigned CRH_VAL( unsigned mode, unsigned conf, unsigned pos)
{
	return CRL_VAL( mode, conf , pos - 8);
}
/* ------------------------------------------------------------------ */
namespace cm3
{
	static inline void nop() { asm volatile ("nop"); }
	static inline void dbarrier() { asm volatile("dsb"); }
}
/* ------------------------------------------------------------------ */
enum dport_dir {
	DPORT_IN,
	DPORT_OUT
};
//LCD RW signal control
inline void rs(bool bit_val)
{
	if(bit_val) LCD_PORT->BSRR = LCD_RS_PIN;
	else LCD_PORT->BRR = LCD_RS_PIN;
}
/* ------------------------------------------------------------------ */
/** GPIO Initialize
 * Initialize bit 15-9 in LCD port
 **/
void gpio_init()
{
	using namespace stm32;	//0x1e7
    //Output mode 10MHz bit15-bit12
	LCD_PORT->CRL &= ~LCD_GPIO_ALL_CRL_MASK;
	LCD_PORT->CRL |= CRL_VAL(GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP,  0 ) |
					 CRL_VAL(GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP,  1 ) |
					 CRL_VAL(GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP,  2 ) |
					 CRL_VAL(GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP,  5 ) |
					 CRL_VAL(GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP,  6 ) |
					 CRL_VAL(GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP,  7 ) ;

   LCD_PORT->CRH &= ~LCD_GPIO_ALL_CRH_MASK;
   LCD_PORT->CRH |= CRH_VAL( GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP, 8);
}
/* ------------------------------------------------------------------ */
//Deinitialize GPIO
void gpio_deinit()
{
	using namespace stm32;	//0x1e7
    //Output mode 10MHz bit15-bit12
	LCD_PORT->CRL &= ~LCD_GPIO_ALL_CRL_MASK;
	LCD_PORT->CRL |= CRL_VAL(GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT,  0 ) |
					 CRL_VAL(GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT,  1 ) |
					 CRL_VAL(GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT,  2 ) |
					 CRL_VAL(GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT,  5 ) |
					 CRL_VAL(GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT,  6 ) |
					 CRL_VAL(GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT,  7 ) ;

   LCD_PORT->CRH &= ~LCD_GPIO_ALL_CRH_MASK;
   LCD_PORT->CRH |= CRH_VAL( GPIO_MODE_INPUT, GPIO_CNF_IN_FLOAT, 8);
}

/* ------------------------------------------------------------------ */
//LCD RW signal control
inline void rw(bool bit_val)
{
	if(bit_val) LCD_PORT->BSRR = LCD_RW_PIN;
	else  LCD_PORT->BRR = LCD_RW_PIN;
	cm3::dbarrier();
}
/* ------------------------------------------------------------------ */
//LCD E signal control
inline void en(bool bit_val)
{
	if(bit_val) LCD_PORT->BSRR = LCD_EN_PIN;
	else  LCD_PORT->BRR = LCD_EN_PIN;
	cm3::dbarrier();
}
/* ------------------------------------------------------------------ */
//Write data to LCD bus
inline void wr_data(uint8_t d_7_4_val)
{
	LCD_PORT->BRR = LCD_DATA_BITS_MASK;
	LCD_PORT->BSRR = ((int)d_7_4_val & 0xf0) << 1;
	cm3::dbarrier();
}
/* ------------------------------------------------------------------ */
//Read data from LCD bus
inline uint8_t rd_data()
{
	uint8_t val =  (LCD_PORT->IDR >> 1)& 0xf0;
	cm3::dbarrier();
	return val;
}

/* ------------------------------------------------------------------ */
//Switch LCD Bus from input to output
inline void dataport_dir(dport_dir out)
{
    using namespace stm32;
    uint32_t tmpL, tmpH;
    tmpH = LCD_PORT->CRH; tmpL = LCD_PORT->CRL;
	if(out==DPORT_OUT)
    {
	   //Output mode 10MHz bit15-bit12
       tmpH &= ~LCD_GPIO_DATA_CRH_MASK;
       tmpH |= CRH_VAL( GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP, 8);
       tmpL &= ~LCD_GPIO_DATA_CRL_MASK;
       tmpL |= CRL_VAL( GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP, 5) |
    		   	   	   	CRL_VAL( GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP, 6) |
    		   	   	   	CRL_VAL( GPIO_MODE_50MHZ, GPIO_CNF_GPIO_PP, 7);
    }
    else
    {

    	//Input mode pull up bit15-bit12
    	tmpH &= ~LCD_GPIO_DATA_CRH_MASK;
    	tmpH |= CRH_VAL( GPIO_MODE_INPUT, GPIO_CNF_IN_PULLUP, 8);
    	tmpL &= ~LCD_GPIO_DATA_CRL_MASK;
    	tmpL |= CRL_VAL( GPIO_MODE_INPUT, GPIO_CNF_IN_PULLUP, 5) |
    	    		   	 CRL_VAL( GPIO_MODE_INPUT, GPIO_CNF_IN_PULLUP, 6) |
    	    		   	 CRL_VAL( GPIO_MODE_INPUT, GPIO_CNF_IN_PULLUP, 7);
        //Pull down configuration
    	LCD_PORT->BSRR = LCD_DATA_BITS_MASK << 16;
    }
	LCD_PORT->CRH = tmpH; LCD_PORT->CRL = tmpL;
	cm3::dbarrier();
}
/* ------------------------------------------------------------------ */
}	//Unnamed namespace end
/* ------------------------------------------------------------------ */
//Destructor
lcd_disp::~lcd_disp()
{
	dbprintf("Destructor called");
	gpio_deinit();
	enable( false );
}
/* ------------------------------------------------------------------ */

void lcd_disp::enable(bool _en)
{
	if(_en)
	{
		//Initialize gpio
		gpio_init();
		en(0);
		//Enable display
		stm32::io_config( LCD_PORT, LCD_ON_PIN, stm32::GPIO_MODE_10MHZ, stm32::GPIO_CNF_GPIO_OD );
		stm32::io_clr( LCD_PORT, LCD_ON_PIN );
	}
	else
	{
		stm32::io_config( LCD_PORT, LCD_ON_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT);
	}
}

/* ------------------------------------------------------------------ */
//Write semi 8 bit on init
void lcd_disp::write8_i(uint8_t addr, uint8_t value )
{
	en(0); rw(0); rs( addr!=0 );
	udelay(1);
	wr_data( value );
	en(1); udelay(1); en(0);
}

/* ------------------------------------------------------------------ */
//Write normal 8 bit as part of
void lcd_disp::write8_n(uint8_t addr, uint8_t value )
{
	en(0); rw(0); rs( addr!=0 );
	wr_data(value);
    en(1);
    udelay(1);
	en(0);
	udelay(1);
	wr_data(value<<4);
	en(1);
	udelay(1);
	en(0);
}

/* ------------------------------------------------------------------ */
//Read 8 bit data
uint8_t lcd_disp::read8( uint8_t addr )
{

	uint8_t rd_val = 0;
	dataport_dir(DPORT_IN);
	en(0); rs( addr!=0 ); rw(1);
	en(1); udelay(1);
	rd_val = rd_data() >> 4;
	udelay(1);
	en(0); udelay(1);
	en(1); udelay(1);
	rd_val |= rd_data();
	udelay(1);
	en(0);
	dataport_dir(DPORT_OUT);
	return rd_val;
}

/* ------------------------------------------------------------------ */
//Delay ms
void lcd_disp::udelay( unsigned period)
{
	period *= 8;
	while(period--)
		cm3::nop();
}

/* ------------------------------------------------------------------ */
//Delay ms
void lcd_disp::mdelay( unsigned period)
{
	tim::timer_wait( period );
}

/* ------------------------------------------------------------------ */
} /* namespace app */
/* ------------------------------------------------------------------ */
