/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * stepup_converter.cpp
 *
 *  Created on: 21-01-2012
 *      Author: lucck
 */
/* ------------------------------------------------------------------ */
#include "stepup_converter.hpp"
#include <stm32tim.h>
#include <system.h>
#include <stm32adc.h>
#include <stm32dma.h>
#include <stm32rcc.h>
#include <dbglog.h>
#include <cstring>
#include <cstdlib>
/* ------------------------------------------------------------------ */
namespace dev
{
/* ------------------------------------------------------------------ */
namespace {
	//ADC divider on/off
	GPIO_TypeDef * const ADC_DIV_EN_PORT = GPIOC;
	const unsigned ADC_DIV_EN_PIN = 1;
	//XDBL pin
	GPIO_TypeDef * const XDBL_PORT = GPIOB;
	const unsigned XDBL_PIN = 14;
	//XPOS pin
	GPIO_TypeDef * const XPOS_PORT = GPIOA;
	const unsigned XPOS_PIN = 8;
	//XPOS pin
	GPIO_TypeDef * const XNEG_PORT = GPIOB;
	const unsigned XNEG_PIN = 13;
	//ADC analog input pin
	GPIO_TypeDef * const ADC1_CH10_PORT = GPIOC;
	const unsigned ADC1_CH10_PIN = 0;
	//Dead time config 500ns
	const int PWM_DEAD_TIME = 12;
	//Constants
	const int MAX_PWM_VALUE = 512;
}
/* ------------------------------------------------------------------ */
namespace {
	//Singleton instance
	stepup_converter *p_conv_ptr;
}

/* ------------------------------------------------------------------ */
inline int stepup_converter::v2pwm(int v)
{
	static const int REG_VOLTAGE = 5000;	//Nominal voltage
	const int e = REG_VOLTAGE - v;
	const int KP = 20;
	if( e == 0 ) return 0;
	else if( std::abs(e) < (int)(REG_VOLTAGE * 0.005) ) return e>0?1:-1;
	else return  e / KP;
}
/* ------------------------------------------------------------------ */
//Default constructor for the singleton
stepup_converter::stepup_converter()
	: m_pwm_value(0)
{
	if( p_conv_ptr ) return;
	std::memset(m_adc_regs, 0, sizeof(m_adc_regs));
	//Enable divider R to GND
	stm32::io_config(ADC_DIV_EN_PORT,ADC_DIV_EN_PIN,stm32::GPIO_MODE_50MHZ,stm32::GPIO_CNF_GPIO_PP);
	stm32::io_clr( ADC_DIV_EN_PORT, ADC_DIV_EN_PIN );
	//Configure xdbl pin To PWM function
	stm32::io_config( XDBL_PORT, XDBL_PIN, stm32::GPIO_MODE_50MHZ, stm32::GPIO_CNF_ALT_PP );
	//Configure timer in pwm mode
	stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb2, RCC_APB2Periph_TIM1, true );
	//Configure timebase
	stm32::tim_timebase_init( TIM1, 0, TIM_CounterMode_Up, MAX_PWM_VALUE, TIM_CKD_DIV1, 0 );
	//Configure OC2N output only XNEG PIN
	stm32::tim_oc_init( TIM1, stm32::tim_cc_chn2, TIM_OCMode_PWM2, 256,
		TIM_OutputState_Disable, TIM_OutputNState_Enable, TIM_OCPolarity_Low,
		TIM_OCNPolarity_Low, TIM_OCIdleState_Reset, TIM_OCNIdleState_Reset );
	//Configure the PWM main output
	//Configure xpos, xneg pin To PWM function
	stm32::io_config( XPOS_PORT, XPOS_PIN, stm32::GPIO_MODE_50MHZ, stm32::GPIO_CNF_ALT_OD );
	stm32::io_config( XNEG_PORT, XNEG_PIN, stm32::GPIO_MODE_50MHZ, stm32::GPIO_CNF_ALT_PP );
	//Configure OC1/IC1N output XPOS, XNEG PIN
	stm32::tim_oc_init( TIM1, stm32::tim_cc_chn1, TIM_OCMode_PWM1, 0,
			TIM_OutputState_Enable, TIM_OutputNState_Enable, TIM_OCPolarity_Low,
			TIM_OCNPolarity_High, TIM_OCIdleState_Reset, TIM_OCNIdleState_Reset );
	//Enable preload register
	stm32::tim_oc_preload_config( TIM1, stm32::tim_cc_chn1, true );
	//Configure dead time
	stm32::tim_bdtr_config( TIM1, PWM_DEAD_TIME, TIM_OSSRState_Disable|TIM_OSSIState_Disable |TIM_LOCKLevel_1
			| TIM_Break_Disable | TIM_BreakPolarity_Low | TIM_AutomaticOutput_Enable );
	//Enable outputs and timer
	stm32::tim_ctrl_pwm_outputs( TIM1, true );
	stm32::tim_cmd( TIM1, true );
	/* ---------- Configure ADC1 and DMA for transfer --------- */
	//Enable ADC1
	stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb2, RCC_APB2Periph_ADC1, true );
	//Configure GPIO as analog inputs
	stm32::io_config( ADC1_CH10_PORT, ADC1_CH10_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_ANALOG );
	//Configure ADC
	stm32::adc_init( ADC1, ADC_Mode_Independent, true, false,
				ADC_ExternalTrigConv_None, ADC_DataAlign_Right, 2 );
	stm32::adc_regular_channel_config( ADC1, ADC_Channel_10,  1,  ADC_SampleTime_55Cycles5 );
	stm32::adc_regular_channel_config( ADC1, ADC_Channel_17,  2,  ADC_SampleTime_55Cycles5 );
	stm32::adc_cmd( ADC1, true );
	stm32::adc_temp_sensor_vref_int_cmd( true );
	stm32::adc_dma_cmd(ADC1, true);
	stm32::adc_reset_calibration(ADC1);
	while(stm32::adc_get_reset_calibration_status(ADC1));
	stm32::adc_start_calibration(ADC1);
	while(stm32::adc_get_calibration_status(ADC1));
	//Configure DMA channel in ADC1
	stm32::dma_enable(stm32::DMACNTR_1);
	stm32::nvic_set_priority( DMA1_Channel1_IRQn, IRQ_PRIO, IRQ_SUBPRIO );
	stm32::nvic_irq_enable( DMA1_Channel1_IRQn, true );
	stm32::dma_channel_config(DMA1_Channel1,
		  DMA_DIR_PeripheralSRC | DMA_PeripheralInc_Disable |
		  DMA_MemoryInc_Enable | DMA_PeripheralDataSize_HalfWord |
		  DMA_MemoryDataSize_HalfWord | DMA_Mode_Circular | DMA_Priority_High | DMA_M2M_Disable,
		   m_adc_regs, &ADC1->DR, 2);
	stm32::dma_irq_enable(DMA1_Channel1,DMA_IT_TC);
	stm32::dma_channel_enable(DMA1_Channel1);
	p_conv_ptr = this;
	dbprintf("Constructor called");
}

/* ------------------------------------------------------------------ */
//Destructor
stepup_converter::~stepup_converter()
{
	//Disable DMA irq
	stm32::nvic_irq_enable( DMA1_Channel1_IRQn, false );
	//Configure DMA channel in ADC1
	stm32::dma_disable(stm32::DMACNTR_1);
	//Disable ADC
	stm32::adc_cmd( ADC1, false );
	stm32::adc_temp_sensor_vref_int_cmd( false );
	stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb2, RCC_APB2Periph_ADC1, false );
	//PWM out 0
	stm32::tim_set_ccr( TIM1, stm32::tim_cc_chn1, 0 );
	//Disable GPIOS
	stm32::io_config(ADC_DIV_EN_PORT,ADC_DIV_EN_PIN, stm32::GPIO_MODE_INPUT,stm32::GPIO_CNF_IN_FLOAT );
	stm32::io_config( XDBL_PORT, XDBL_PIN, stm32::GPIO_MODE_INPUT,stm32::GPIO_CNF_IN_FLOAT );
	stm32::io_config( XPOS_PORT, XPOS_PIN, stm32::GPIO_MODE_INPUT,stm32::GPIO_CNF_IN_FLOAT );
	stm32::io_config( XNEG_PORT, XNEG_PIN, stm32::GPIO_MODE_INPUT,stm32::GPIO_CNF_IN_FLOAT );
	stm32::io_config( ADC1_CH10_PORT, ADC1_CH10_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT );
	//Disable TIM1
	stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb2, RCC_APB2Periph_TIM1, false );
	//Conv ptr
	p_conv_ptr = NULL;
	dbprintf("Destructor called");
}
/* ------------------------------------------------------------------ */
//Start conversion
inline void stepup_converter::start_conv()
{
	stm32::adc_software_start_conv_cmd( ADC1, true );
}
/* ------------------------------------------------------------------ */
//Calculate PWM output
inline void stepup_converter::calculate_pwm_out()
{
	static const int MAX_ALLOWED_PWM = 0.66 * MAX_PWM_VALUE;		//Max allowed pwm
	m_pwm_value += v2pwm( get_Vzas());
	if( m_pwm_value > MAX_ALLOWED_PWM ) m_pwm_value = MAX_ALLOWED_PWM;
	else if( m_pwm_value < 0 ) m_pwm_value = 0;
	//Calculate output
	stm32::tim_set_ccr( TIM1, stm32::tim_cc_chn1, m_pwm_value );
}

/* ------------------------------------------------------------------ */
namespace tim {
namespace _internal {
	void stepup_converter_tick_handler()
	{
		if( p_conv_ptr )
			p_conv_ptr->start_conv();
	}
}}
/* ------------------------------------------------------------------ */
namespace _internal {
extern "C" {
	void dma1_channel1_isr_vector(void) __attribute__((interrupt));
	void dma1_channel1_isr_vector(void)
	{
		stm32::dma_clear_flag( DMA1_FLAG_TC1);
		if( p_conv_ptr )
			p_conv_ptr->calculate_pwm_out();
	}
}}

/* ------------------------------------------------------------------ */
} /* namespace app2 */
/* ------------------------------------------------------------------ */
