/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * theapplication.cpp
 *
 *  Created on: 05-01-2012
 *      Author: lucck
 */

/* ------------------------------------------------------------------ */
#include "theapplication.h"
#include "stepup_converter.hpp"
#include "sys_time.hpp"
#include "config.h"
#include <dbglog.h>
#include <stdlib.h>
#include <cmath>
#include <stm32bkp.h>

#ifdef ENABLE_SERVICE_CALIBRATION_MENU
#include <utils.h>
#include <stm32fmc.hpp>
#include <stm32rcc.h>
#endif

/* ------------------------------------------------------------------ */
namespace app
{
/* ------------------------------------------------------------------ */
namespace
{
	//Enum incrementer
	template <typename T> inline T& operator++( T& v )
	{
		v = static_cast<T>(static_cast<int>(v) + 1);
		return v;
	}
	//Enum decrementer
	template <typename T> inline T& operator--( T& v )
	{
		v = static_cast<T>(static_cast<int>(v) - 1);
		return v;
	}
	//Calculate pow10
	inline int get_pow10( int value )
	{
	    int ret;
	    for(ret = 0; value; value/=10, ret++);
	    return ret-1;
	}
	const char* const type_measure_str[] =
	{
	    "C=",	//C calculate
		"L=",	//L calculate
		"FH=",	//Frequency hi
        "FL="	//Frequency low
    };
	const char* const final_units_str[] =
	{
		"F",		//Farad
		"H",		//Henr
        "Hz",		//Hertz
		"Hz"		//Hertz
	};
	const real_t SI_units[] =
	{
		1.0E12,		//Pik
		1.0E9,		//Nano
		1.0E6,		//Mikro
		1.0,		//No unit
		1.0E3,		//Mili
		1.0E-3,		//Kilo
		1.0E-6,		//Mega
	};
	const char* const SI_units_str[] =
	{
		"p",
		"n",
		"\xE4",	//Micro on hd44780
		" ",
		"m",
		"k",
		"M"
	};
	//For all measurements
	const int SI_prec_allM[] =
	{
		10,			// Piko prec
		1000,		// Nano prec
		1000,		//Micro prec
		1,			//Normal
		1000,		//Mili prec
		1000,		//Kilo prec
		1000 * 1000	//Mega prec
	};
	//for inductors only
	const int SI_prec_LM[] =
	{
		10,			// Piko prec
		1,			// Nano prec
		1000,		//Micro prec
		1,			//Normal
		1000,		//Mili prec
		1000,		//Kilo prec
		1000 * 1000	//Mega prec
	};
	const char* const modes_str[] =
	{
		"C ",	//C calculate
		"L ",	//L calculate
        "FH",	//Frequency low
		"FL",	//Frequency hi
	};
}
/* ------------------------------------------------------------------ */
//Constructor
the_application::the_application()
	: m_mode( e_mode_disp_C ),  m_curr_disp(&m_lc_meter),
	  m_last_bottom_refresh_time(-1), m_last_used_time(0)
{
	//Enable the lcd device
	m_lcd.enable( true );
	m_curr_disp->enable( true );
}


/* ------------------------------------------------------------------ */
//Display the value
void the_application::display_measured_value( real_t value )
{
	using namespace fnd::lcd;
	m_lcd << pos(1,1) << type_measure_str[m_mode];
	if( value == 0.0 )
	{
		m_lcd << "0";
	}
	else if( std::isnan( value ) )
	{
		m_lcd << "******";
	}
	else
	{
		const int * const SI_prec = (m_mode==e_mode_disp_L)?(SI_prec_LM):(SI_prec_allM);
		size_t u;
		for(u=0; u<sizeof(SI_units)/sizeof(SI_units[0]); ++u)
		{
			real_t m_value = fabs(value * SI_units[u]);
			if( m_value >= 1.0  && m_value < 1000.0 )
				break;
		}
		//Check if we are able to display value
		if( u ==sizeof(SI_units)/sizeof(SI_units[0]))
		{
			for(u=0; u<sizeof(SI_units)/sizeof(SI_units[0]); ++u)
			{
				real_t m_value = fabs(value * SI_units[u]);
				if (m_value >= 1.0/SI_prec[u] ) break;
			}
		}
		if( u ==sizeof(SI_units)/sizeof(SI_units[0]))
		{
			m_lcd << "0";
		}
		else
		{
			const int ivalue = static_cast<int>(value * SI_units[u] * static_cast<real_t>(SI_prec[u]) + 0.5);
			const int zeros = get_pow10(SI_prec[u]);
			m_lcd << ivalue/SI_prec[u];
			if(zeros > 0)
				m_lcd << "." << lfmt( abs(ivalue%SI_prec[u]),zeros,'0' );
			m_lcd << SI_units_str[u];
		}
	}
	m_lcd << final_units_str[m_mode];
	m_lcd.erase(12);
}
/* ------------------------------------------------------------------ */
//Display bottom line (normal battery status)
void the_application::display_battery_status( )
{
	using namespace dev::tim;
	using namespace fnd::lcd;
	static const int C_bottom_refresh_time = 2000;
	static const int C_battery_bar_width = 50;
	static const int C_min_bat_level = 2000;
	static const int C_max_bat_level = 3100;
	if( timer_elapsed( m_last_bottom_refresh_time, C_bottom_refresh_time ) )
	{
		const int bar_pos = ((m_power_conv.get_Vbat() - C_min_bat_level) * C_battery_bar_width)
				/ (C_max_bat_level - C_min_bat_level);
		m_lcd << pos(1,2) << "Bat [";
		m_lcd.progress_bar( 6, 2, bar_pos, C_battery_bar_width );
		m_lcd << pos(16,2) << "]";
		m_last_bottom_refresh_time = read_tick();
	}
}
/* ------------------------------------------------------------------ */
//Display calibration screen
void the_application::display_calibration_screen( int percent, int value )
{
	using namespace fnd::lcd;
	static const int C_calib_bar_width = 45;
	static const int C_range_100_percent = 100;
	const int bar_pos = ( percent * C_calib_bar_width ) / C_range_100_percent;
	m_lcd << pos(1,1) << "Calib " <<  modes_str[ m_mode ] << value;
	m_lcd.erase( 10 );
	m_lcd << pos( 1,2 ) << "Wait [" << pos(16,2) << "]";
	m_lcd.progress_bar( 7, 2, bar_pos, C_calib_bar_width );
}
/* ------------------------------------------------------------------ */
//On change mode
void the_application::on_change_mode()
{
	if(	 m_lc_meter.get_mode()==lc_measure::mode_calibrate_L
		|| m_lc_meter.get_mode()==lc_measure::mode_calibrate_C )	//If app is during calibration block it
	{
		if( m_mode == e_mode_disp_C ) m_lc_meter.set_mode( lc_measure::mode_C );
		else if( m_mode == e_mode_disp_L ) m_lc_meter.set_mode( lc_measure::mode_L );
	}
	if( m_mode == e_mode_disp_C || m_mode == e_mode_disp_L )
	{
		if(m_curr_disp != &m_lc_meter)	//Switch power states if required
		{
			m_f_meter.enable( false );
			m_lc_meter.enable( true );
			dbprintf("Disable F Enable LC");
		}
		m_curr_disp = &m_lc_meter;
	}
	else if( m_mode == e_mode_disp_F_hi || m_mode == e_mode_disp_F_low )
	{
		if( m_curr_disp != &m_f_meter )	//Switch power states if required
		{
			m_f_meter.enable( true );
			m_lc_meter.enable( false );
			dbprintf("Enable F disable LC");
		}
		m_curr_disp = &m_f_meter;
	}
	//Set working objects mode according to current menu
	switch(m_mode)
	{
	case e_mode_disp_C: 	m_lc_meter.set_mode( lc_measure::mode_C );				break;
	case e_mode_disp_L:		m_lc_meter.set_mode( lc_measure::mode_L ); 				break;
	case e_mode_disp_F_low: m_f_meter.set_mode( freqency_measure::mode_direct ); 	break;
	case e_mode_disp_F_hi:  m_f_meter.set_mode( freqency_measure::mode_divide_64 ); break;
	default:																		break;
	}
}
/* ------------------------------------------------------------------ */
//On start calibration
void the_application::on_start_calibration()
{
	if( m_lc_meter.get_mode()!=lc_measure::mode_calibrate_L &&
	    m_lc_meter.get_mode()!=lc_measure::mode_calibrate_C
	  )	//If not in calib mode start cal
	{
		dbprintf(" Calibration pressed ");
		if( m_mode == e_mode_disp_L)
		{
			m_lc_meter.set_mode( lc_measure::mode_calibrate_L );
		}
		else if( m_mode == e_mode_disp_C)
		{
			m_lc_meter.set_mode( lc_measure::mode_calibrate_C );
		}
	}
}
/* ------------------------------------------------------------------ */
//On display data
void the_application::do_process_data( i_measure &measure )
{
	using namespace dev;
	real_t value;
	const int ret_code =  measure(value);
	//if (ret_code != 2) dbprintf("Ret code is %d", ret_code);
	if(ret_code == i_measure::RESULT_NO_NEWDATA)
		return;
	//Check the shutdown time
	if( !measure.is_idle() )
		m_last_used_time = tim::read_tick();
	if( ret_code == i_measure::RESULT_OK || ret_code == i_measure::RESULT_INVALID )
	{
		if(ret_code==i_measure::RESULT_INVALID) value = NAN;
		display_measured_value( value );
		//Display bottom line (voltage)
		display_battery_status();
	}
	else if( ret_code == i_measure::RESULT_NO_CALIBRATION )
	{
		using namespace fnd::lcd;
		m_lcd << pos(1,1) << modes_str[ m_mode ] << " no calibrated";
		m_lcd << pos(1,2) << "Please Calibrate";
	}
    else if( ret_code == i_measure::RESULT_INVALID_CALIBRATION )
    {
        using namespace fnd::lcd;
        m_lcd << pos(1,1) << modes_str[ m_mode ] << " Calib failure";
        m_lcd << pos(1,2) << "Short terminals ";
    }
	//If calibration screen is required
	else if( ret_code >= i_measure::RESULT_CALIBRATE_0PERCENT &&
			ret_code <= i_measure::RESULT_CALIBRATE_100PERCENT )
	{
		display_calibration_screen( ret_code - i_measure::RESULT_CALIBRATE_0PERCENT , value );
	}
}
/* ------------------------------------------------------------------ */
//Increment mode
void the_application::increment_mode()
{
	//Only if not in calibrate mode change value
	if(	 m_lc_meter.get_mode()!=lc_measure::mode_calibrate_L
		 && m_lc_meter.get_mode()!=lc_measure::mode_calibrate_C )
	{
		if(++m_mode >= e_mode_disp_last ) m_mode = e_mode_disp_C;				//forward
		m_last_bottom_refresh_time = 0;			//Force refresh if menu changed
	}
}
/* ------------------------------------------------------------------ */
//Decrement mode
void the_application::decrement_mode()
{
	//Only if not in calibrate mode change value
	if(	 m_lc_meter.get_mode()!=lc_measure::mode_calibrate_L
		 && m_lc_meter.get_mode()!=lc_measure::mode_calibrate_C )
	{
		if(--m_mode <= e_mode_first ) m_mode = e_mode_disp_F_low;		//forward
		m_last_bottom_refresh_time = 0;					//Force refresh if menu changed
	}
}
/* ------------------------------------------------------------------ */
#ifdef ENABLE_SERVICE_CALIBRATION_MENU
//Set calibration value
void the_application::service_mode_calibration()
{
	using namespace dev;
	using namespace fnd::lcd;
	const int info_timeout = 4000;
	m_lcd << pos(1,1) << "Ver" << config::app_version << " Calibr.   ";
	m_lcd.cursor( true );
	char e_buf[16];
	const size_t e_buf_len = 7;
	const size_t dot_pos = 4;
	const int C_int_val = m_lc_meter.get_Cref_value() * 1E14;
	fnd::fnd_itoa(  e_buf, C_int_val /100 , 4, '0' );
	e_buf[dot_pos] = '.';
	fnd::fnd_itoa( &e_buf[dot_pos+1], C_int_val %100 , 2, '0' );
	size_t cpos=0;
	for(int key=0x100; ; key = m_kbd.get())
	{
		if( key == keyboard::KEY_OFF)
		{
			if(++cpos>=e_buf_len) cpos = 0;
			if(cpos==dot_pos) ++cpos;
		}
		else if( key == keyboard::KEY_ON_MODE)
		{
			if( ++e_buf[cpos] > '9') e_buf[cpos] = '0';
		}
		if( key )
		{
			m_lcd << pos(1,2) << "CCal=" << e_buf << "pF  ";
			m_lcd.setpos( cpos+6, 2);
		}
		if(  key == (keyboard::KEY_OFF|keyboard::KEY_RPT ) )
		{
			m_lcd << pos(1,1) << "Write calibrat. "
				  << pos(1,2) << "Aborted...      ";
			tim::timer_wait(  info_timeout );
			break;
 		}
		else if(  key == (keyboard::KEY_ON_MODE|keyboard::KEY_RPT ) )
		{
			using namespace stm32;
			int errno = stm32fmc::ERRNO_COMPLETE;
			const int C_invalid_conversion = 1;
			e_buf[dot_pos] = '\0';
			const int write_value = fnd::fnd_atoui( e_buf )*100 + fnd::fnd_atoui( &e_buf[dot_pos+1] );
			dbprintf("Write CCal value %d", write_value );
			stm32::rcc_hsi_cmd( true );
			stm32fmc flash_ctrl;
			stm32::rcc_hsi_cmd( false );
			do
			{
				if( write_value < lc_measure::Cref_int_min || write_value > lc_measure::Cref_int_max )
				{
					errno = C_invalid_conversion;
					dbprintf("Set invalid value");
					break;
				}
				if((errno = flash_ctrl.page_erase(flash_ctrl.num_pages()-1))!=stm32fmc::ERRNO_COMPLETE)
					break;
				if( ( errno = flash_ctrl.write_word(
							flash_ctrl.num_pages()-1, flash_ctrl.page_size()-sizeof(write_value), write_value )
					)!=stm32fmc::ERRNO_COMPLETE)
				   break;
			}
			while(0);
			m_lcd << pos(1,1) << "Write calibrat. ";
			if(errno == stm32fmc::ERRNO_COMPLETE )
				m_lcd << pos(1,2) << "Completed...    ";
			else
				m_lcd << pos(1,2) << "failed err=" << errno << "     ";
			tim::timer_wait(  info_timeout );
			break;
		}
		//Sleep
		goto_sleep();
	}
	//Flush kbd disable cursor
	m_kbd.get();
	m_lcd.cursor( false );
}
#endif
/* ------------------------------------------------------------------ */
//About screen
void the_application::about_screen()
{
	using namespace fnd::lcd;
	using namespace dev;
	const int C_about_timeout = 2000;
	m_lcd << pos(1,1) << "LCFMeter" << config::app_version << " BoFF";
	m_lcd << pos(1,2) << "www.boff.pl 2011";
	tim::timer_wait( C_about_timeout  );
}
/* ------------------------------------------------------------------ */
//The application main loop
void the_application::run()
{
	using namespace dev;
	//Display about screen on startup
	about_screen();
	//Before main loop start calibration (setup, the key)
#ifdef ENABLE_SERVICE_CALIBRATION_MENU
	if( m_kbd.wait_for_release_all() == (keyboard::KEY_OFF| keyboard::KEY_RPT ))
	{
		service_mode_calibration();
	}
#else
	 m_kbd.wait_for_release_all();
#endif
	for(int key=0; key!= (keyboard::KEY_OFF|keyboard::KEY_RPT); key = m_kbd.get())
	{

		//Used timeout update
		if( key ) m_last_used_time = tim::read_tick();
		if( tim::timer_elapsed(m_last_used_time, C_shutdown_timeout ) ) break;

		if( key == keyboard::KEY_ON_MODE )
		{
			//Increment to the next mode
			increment_mode();
			//Change mode
			on_change_mode();
		}
		else if( key == keyboard::KEY_OFF)
		{
			//Increment to the next mode
			decrement_mode();
			//Change mode
			on_change_mode();
		}
		else if( key == (keyboard::KEY_ON_MODE|keyboard::KEY_RPT ) )
		{
			//Start calibration
			on_start_calibration();
		}
		//Do process data
		do_process_data( *m_curr_disp );
		//Wait for interrupt
		goto_sleep();
	}
}
/* ------------------------------------------------------------------ */
} /* namespace app */
/* ------------------------------------------------------------------ */
