/* 
    LCFMETER - Firmware for measure L, C, F using STM32F100 device
    Copyright (C) 2012  Lucjan Bryndza <lucck@boff.pl>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. 
*/

/*
 * =====================================================================================
 *
 *       Filename:  tim_freq_measure.cpp
 *
 *    Description:  timer freq measure
 *
 *        Version:  1.0
 *        Created:  31.01.2012 13:35:39
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Lucjan Bryndza (lb), lucck@boff.pl
 *   Organization:  
 *
 * =====================================================================================
 */

/* #####   HEADER FILE INCLUDES   ################################################### */

#include "tim_freq_measure.hpp"
#include "config.h"
#include <stm32tim.h>
#include <stm32rcc.h>
#include <dbglog.h>
/* ------------------------------------------------------------------ */
namespace dev {
/* ------------------------------------------------------------------ */
//Internal namespace
namespace
{
	//Freq input 1 pin
	GPIO_TypeDef * const FREQ1_PORT = GPIOA;
	const int FREQ1_PIN = 0;
	//Freq input 2 pin
	GPIO_TypeDef * const FREQ2_PORT = GPIOD;
	const int FREQ2_PIN = 2;
	tim_freq_measure * p_tim3;
	tim_freq_measure * p_tim2;
}
/* ------------------------------------------------------------------ */
//Timer freq measure
tim_freq_measure::tim_freq_measure( TIM_TypeDef * tim )
	: m_tim(tim), m_hi_cnt(0), m_capture(0)
{
	/* ***************** Configure gate timer (TIM15) ***************** */
	if( !p_tim2 && !p_tim3 )	//Not initialized devs
	{
		//Enable timer
		stm32::rcc_periph_clock_cmd(stm32::rcc_clk_apb2, RCC_APB2Periph_TIM15, true );
		//Configure time base
		stm32::tim_timebase_init(TIM15, config::PCLK2_HZ / BASE_PERIOD - 1,
				TIM_CounterMode_Up, GATE_TIME_MS + GATE_NOT_ACTIVE_MS, 0, 0);
		//Select master slave mode
		stm32::tim_select_master_slave_mode( TIM15, TIM_MasterSlaveMode_Enable );
		//Configure OC1 as gate for slave timer
		stm32::tim_oc_init( TIM15, stm32::tim_cc_chn1, TIM_OCMode_PWM1, GATE_TIME_MS,
				TIM_OutputState_Disable, TIM_OutputNState_Disable, TIM_OCPolarity_Low,
				TIM_OCNPolarity_Low, TIM_OCIdleState_Reset, TIM_OCNIdleState_Reset );
		//Select trigger output
		stm32::tim_select_output_trigger( TIM15, TIM_TRGOSource_OC1Ref);
		//Configure NVIC interrupt
		stm32::nvic_set_priority( TIM1_BRK_TIM15_IRQn, IRQ_PRIO, IRQ_SUB );
		stm32::nvic_irq_enable( TIM1_BRK_TIM15_IRQn, true );
		stm32::tim_it_config( TIM15, TIM_IT_CC1, true );
		//Enable the timer
		stm32::tim_cmd( TIM15, true);
	}
	/* ***************** Configure TIM2 or TIM3 ***************** */
	if( m_tim == TIM2 )
	{
		if( p_tim2 ) return;
		//Enable tim2 clock
		stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb1, RCC_APB1Periph_TIM2, true );
		stm32::io_config( FREQ1_PORT, FREQ1_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT );
		p_tim2 = this;
	}
	else if( m_tim == TIM3)
	{
		if( p_tim3 ) return;
		//Enable tim3 clock
		stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb1, RCC_APB1Periph_TIM3, true );
		stm32::io_config( FREQ2_PORT, FREQ2_PIN, stm32::GPIO_MODE_INPUT, stm32::GPIO_CNF_IN_FLOAT );
		p_tim3 = this;
	}
	else
	{
		return;
	}
	//Initialize the timebase
	stm32::tim_timebase_init(m_tim, 0, TIM_CounterMode_Up, 65535, 0, 0 );
	//Select slave mode
	stm32::tim_select_slave_mode(m_tim, TIM_SlaveMode_Gated );
	//Select input trigger in slave mode
	stm32::tim_select_input_trigger( m_tim, (m_tim==TIM2)?(TIM_TS_ITR1):(TIM_TS_ITR2) );
	//Select external trigger
	stm32::tim_etr_clock_mode2_config( m_tim, TIM_ExtTRGPSC_OFF, TIM_ExtTRGPolarity_NonInverted, 0 );
	//Configure Capture
	stm32::tim_it_config( m_tim, TIM_IT_Update, true );
	//Configure NVIC interrupt
	stm32::nvic_set_priority( m_tim==TIM2?TIM2_IRQn:TIM3_IRQn, IRQ_PRIO, IRQ_SUB );
	stm32::nvic_irq_enable( m_tim==TIM2?TIM2_IRQn:TIM3_IRQn, true );
	//enable timer
	stm32::tim_cmd( m_tim, true );
}
/* ------------------------------------------------------------------ */
//Destructor
tim_freq_measure::~tim_freq_measure()
{
	dbprintf("Destructor called");
	if( p_tim2 && m_tim == TIM2 )
	{
		//Enable tim2 clock
		stm32::tim_cmd( TIM2, false );
		stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb1, RCC_APB1Periph_TIM2, false );
		p_tim2 = NULL;
	}
	else if( p_tim3 && m_tim == TIM3)
	{
		//Enable tim3 clock
		stm32::tim_cmd( TIM3, false );
		stm32::rcc_periph_clock_cmd( stm32::rcc_clk_apb1, RCC_APB1Periph_TIM3, false );
		p_tim3 = NULL;
	}
	if( !p_tim2 && !p_tim3 )
	{
		//Disable timer
		stm32::tim_cmd( TIM15, false );
		stm32::rcc_periph_clock_cmd(stm32::rcc_clk_apb2, RCC_APB2Periph_TIM15, false );
	}
}
/* ------------------------------------------------------------------ */
//Handle interrupt
void tim_freq_measure::handle_overflow()
{
	if( stm32::tim_get_it_status(m_tim, TIM_IT_Update ) )
	{
		stm32::tim_clear_it_pending_bit( m_tim, TIM_IT_Update );
		m_hi_cnt++;
	}
}
/* ------------------------------------------------------------------ */
//Handle measure finished
void tim_freq_measure::handle_measure_finished()
{
	uint32_t capture =  (static_cast<uint32_t>(m_hi_cnt)<<16)| stm32::tim_get_counter( m_tim );
	stm32::atomic_try_write_word( &m_capture, capture );
	m_hi_cnt = 0;
	stm32::tim_set_counter( m_tim, 0 );
}
/* ------------------------------------------------------------------ */
namespace _internal {
extern "C" {
	//Timer2 isr vector
	void tim2_isr_vector(void)
	{
		if( p_tim2 )
			p_tim2->handle_overflow();
	}
	//Timer3 isr vector
	void tim3_isr_vector(void)
	{
		if( p_tim3 )
			p_tim3->handle_overflow();
	}
	//Timer 15 event for end of counting
	void tim1_brk_tim15_isr_vector(void)
	{
		if( stm32::tim_get_it_status(TIM15, TIM_IT_CC1 ) )
		{
			stm32::tim_clear_it_pending_bit( TIM15, TIM_IT_CC1 );
			if( p_tim2 )
				p_tim2->handle_measure_finished();
			if ( p_tim3 )
				p_tim3->handle_measure_finished();
		}
	}
}}
/* ------------------------------------------------------------------ */
}		/* -----  end of namespace dev  ----- */

